package ctbrec.recorder.postprocessing;

import ctbrec.Config;
import ctbrec.Recording;
import ctbrec.StringUtil;
import ctbrec.variableexpansion.ModelVariableExpander;
import ctbrec.variableexpansion.ParserVisitor;
import ctbrec.variableexpansion.VariableExpansionException;
import ctbrec.variableexpansion.antlr.PostProcessingLexer;
import ctbrec.variableexpansion.antlr.PostProcessingParser;
import ctbrec.variableexpansion.functions.AntlrSyntacErrorAdapter;
import lombok.extern.slf4j.Slf4j;
import org.antlr.v4.runtime.CharStream;
import org.antlr.v4.runtime.CharStreams;
import org.antlr.v4.runtime.CommonTokenStream;
import org.antlr.v4.runtime.tree.ParseTree;

import java.io.IOException;
import java.io.StringReader;
import java.time.Instant;
import java.time.ZoneId;
import java.time.ZoneOffset;
import java.util.HashMap;
import java.util.Map;
import java.util.Optional;

import static java.util.Optional.ofNullable;

@Slf4j
public abstract class AbstractPlaceholderAwarePostProcessor extends AbstractPostProcessor {

    public String fillInPlaceHolders(String input, PostProcessingContext ctx) {
        return this.fillInPlaceHolders(input, ctx, null);
    }

    public String fillInPlaceHolders(String input, PostProcessingContext ctx, AntlrSyntacErrorAdapter errorListener) {
        Recording rec = ctx.getRecording();
        Config config = ctx.getConfig();

        ModelVariableExpander modelExpander = new ModelVariableExpander(rec.getModel(), config, ctx.getRecorder());
        Map<String, Optional<Object>> variables = new HashMap<>(modelExpander.getPlaceholderValueSuppliers());
        variables.put("recordingNotes", getSanitizedRecordingNotes(rec));
        variables.put("fileSuffix", getFileSuffix(rec));
        variables.put("recordingsDir", Optional.of(config.getSettings().recordingsDir));
        variables.put("absolutePath", Optional.of(rec.getPostProcessedFile().getAbsolutePath()));
        variables.put("absoluteParentPath", Optional.of(rec.getPostProcessedFile().getParentFile().getAbsolutePath()));
        variables.put("epochSecond", ofNullable(rec.getStartDate()).map(Instant::getEpochSecond).map(l -> Long.toString(l))); // NOSONAR
        variables.put("utcDateTime", getUtcDateTime(rec));
        variables.put("localDateTime", getLocalDateTime(rec));

        return fillInPlaceHolders(input, variables, errorListener);
    }

    private String fillInPlaceHolders(String input, Map<String, Optional<Object>> variables, AntlrSyntacErrorAdapter errorListener) {
        try (StringReader reader = new StringReader(input)) {
            CharStream s = CharStreams.fromReader(reader);
            PostProcessingLexer lexer = new PostProcessingLexer(s);
            CommonTokenStream tokens = new CommonTokenStream(lexer);
            PostProcessingParser parser = new PostProcessingParser(tokens);
            Optional.ofNullable(errorListener).ifPresent(parser::addErrorListener);
            ParseTree parseTree = parser.line();
            ParserVisitor visitor = new ParserVisitor(variables);
            return visitor.visit(parseTree);
        } catch (IOException e) {
            throw new VariableExpansionException("Couldn't replace placeholders", e);
        }
    }

    protected Optional<Object> getUtcDateTime(Recording rec) {
        return Optional.ofNullable(rec)
                .map(Recording::getStartDate)
                .map(i -> i.atZone(ZoneOffset.UTC));
    }

    protected Optional<Object> getLocalDateTime(Recording rec) {
        return Optional.ofNullable(rec)
                .map(Recording::getStartDate)
                .map(i -> i.atZone(ZoneId.systemDefault()));
    }

    private Optional<Object> getFileSuffix(Recording rec) {
        if (rec.isSingleFile()) {
            String filename = rec.getPostProcessedFile().getName();
            return Optional.of(filename.substring(filename.lastIndexOf('.') + 1));
        } else {
            return Optional.empty();
        }
    }

    private Optional<Object> getSanitizedRecordingNotes(Recording rec) {
        Optional<String> notes = ofNullable(rec.getNote());
        return notes.map(StringUtil::sanitize);
    }
}
